==> sisu_7.2.0.orig.tar.xz: sisu-7.2.0/lib/sisu/json.rb <==
# encoding: utf-8
=begin

* Name: SiSU

** Description: documents, structuring, processing, publishing, search
*** Description: json output logic, flow

** Author: Ralph Amissah
  [ralph@amissah.com]
  [ralph.amissah@gmail.com]

** Copyright: (C) 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006,
  2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016 Ralph Amissah,
  All Rights Reserved.

** License: GPL 3 or later:

  SiSU, a framework for document structuring, publishing and search

  Copyright (C) Ralph Amissah

  This program is free software: you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program. If not, see [http://www.gnu.org/licenses/].

  If you have Internet connection, the latest version of the GPL should be
  available at these locations:
  [http://www.fsf.org/licensing/licenses/gpl.html]
  [http://www.gnu.org/licenses/gpl.html]

** SiSU uses:
  * Standard SiSU markup syntax,
  * Standard SiSU meta-markup syntax, and the
  * Standard SiSU object citation numbering and system

** Hompages:
  [http://www.jus.uio.no/sisu]
  [http://www.sisudoc.org]

** Git
  [http://git.sisudoc.org/gitweb/?p=code/sisu.git;a=summary]
  [http://git.sisudoc.org/gitweb/?p=code/sisu.git;a=blob;f=lib/sisu/json.rb;hb=HEAD]

=end
module SiSU_JSON
  require_relative 'se_hub_particulars'                 # se_hub_particulars.rb
    include SiSU_Particulars
  require_relative 'se'                                 # se.rb
    include SiSU_Env
  require_relative 'json_shared'                        # json_shared.rb
    include SiSU_JSON_Munge
  require_relative 'json_format'                        # json_format.rb
    include SiSU_JSON_Format
  require_relative 'json_persist'                       # json_persist.rb
  require_relative 'shared_metadata'                    # shared_metadata.rb
  @@alt_id_count=0
  @@tablefoot=''
  class Source
    def initialize(opt)
      @opt=opt
      @particulars=SiSU_Particulars::CombinedSingleton.instance.get_all(opt)
    end
    def read
      begin
        @env,@md,@ao_array=@particulars.env,@particulars.md,@particulars.ao_array
        unless @opt.act[:quiet][:set]==:on
          tool=if (@opt.act[:verbose_plus][:set]==:on \
          || @opt.act[:maintenance][:set]==:on)
            @env.program.web_browser +
            ' file://' +
            @md.file.output_path.json.dir + '/' +
            @md.file.base_filename.json
          elsif @opt.act[:verbose][:set]==:on
            @env.program.web_browser +
            ' file://' +
            @md.file.output_path.json.dir + '/' +
            @md.file.base_filename.json
          else "[#{@opt.f_pth[:lng_is]}] #{@opt.fno}"
          end
          (@opt.act[:verbose][:set]==:on \
          || @opt.act[:verbose_plus][:set]==:on \
          || @opt.act[:maintenance][:set]==:on) \
          ? SiSU_Screen::Ansi.new(
              @opt.act[:color_state][:set],
              'JSON',
              tool
            ).green_hi_blue
          : SiSU_Screen::Ansi.new(
              @opt.act[:color_state][:set],
              'JSON',
              tool
            ).green_title_hi
          if (@opt.act[:verbose_plus][:set]==:on \
          || @opt.act[:maintenance][:set]==:on)
            SiSU_Screen::Ansi.new(
              @opt.act[:color_state][:set],
              @opt.fns,
                '/' + @md.file.output_path.json.dir +
                '/' + @md.file.base_filename.json
            ).flow
          end
        end
        SiSU_JSON::Source::Songsheet.new(@particulars).song
      rescue
        SiSU_Errors::Rescued.new($!,$@,@opt.selections.str,@opt.fns).location do
          __LINE__.to_s + ':' + __FILE__
        end
      ensure
        SiSU_Env::CreateSite.new(@opt).cp_css
        Dir.chdir(@opt.f_pth[:pth])
      end
    end
    private
    class Songsheet
      def initialize(particulars)
        @env,@md,@ao_array,@particulars=
          particulars.env,particulars.md,particulars.ao_array,particulars
        @file=SiSU_Env::FileOp.new(@md)
      end
      def song
        begin
          SiSU_JSON::Source::Scroll.new(@particulars).songsheet
        rescue
          SiSU_Errors::Rescued.new($!,$@,@md.opt.selections.str,@md.fns).location do
            __LINE__.to_s + ':' + __FILE__
          end
        ensure
        end
      end
    end
    class Scroll
      require_relative 'json_shared'                   # json_shared.rb #check already called
      require_relative 'txt_shared'                     # txt_shared.rb
        include SiSU_TextUtils
      require_relative 'css'                            # css.rb
      def initialize(particulars)
        @env,@md,@ao_array=particulars.env,particulars.md,particulars.ao_array
        @tab="\t"
        @trans=SiSU_JSON_Munge::Trans.new(@md)
        @sys=SiSU_Env::SystemCall.new
        @per=SiSU_JSON_Persist::Persist.new
      end
      def songsheet
        begin
          pre
          @data=markup(@ao_array)
          post
          publish
        ensure
          SiSU_JSON_Persist::Persist.new.persist_init
        end
      end
    protected
      def embedded_endnotes(dob='')
        dob.obj=dob.obj.gsub(/#{Mx[:en_a_o]}(\d+)\s+(.+?)#{Mx[:en_a_c]}/,
            '<endnote><number>\1</number><note>\2</note></endnote> ').
          gsub(/#{Mx[:en_b_o]}([*+]\d+)\s+(.+?)#{Mx[:en_b_c]}/,
            '<endnote><symbol>\1</symbol><note>\2</note></endnote> ').
          gsub(/#{Mx[:en_a_o]}([*+]+)\s+(.+?)#{Mx[:en_a_c]}/,
            '<endnote><symbol>\1</symbol><note>\2</note></endnote> ')
      end
      def extract_endnotes(dob='')
        notes=dob.obj.scan(/(?:#{Mx[:en_a_o]}|#{Mx[:en_b_o]})([\d*+]+\s+.+?)(?:#{Mx[:en_a_c]}|#{Mx[:en_b_c]})/)
        notes.flatten.each do |e|
          s=e.to_s
          util=SiSU_JSONutils::Clean.new(s)
          wrap=util.line_json_clean
          wrap=wrap.gsub(/^(\d+)\s+(.+?)\s*\Z/m, <<-WOK
\\n[\\1.] \\2
              WOK
            ).
            gsub(/^([*+]\d+)\s+(.+?)\s*\Z/m, <<-WOK
\\n[\\1.] \\2
              WOK
            ).
            gsub(/^([*+]+)\s+(.+?)\s*\Z/m, <<-WOK
\\n[\\1.] \\2
              WOK
            ).strip
#KEEP alternative presentation of endnotes
#        wrap=wrap.gsub(/^(\d+)\s+(.+?)\s*\Z/m, <<WOK
##{Ax[:tab]*1}<p class="endnote" notenumber="\\1">
##{Ax[:tab]*2}\\1. \\2
##{Ax[:tab]*1}</p>
#WOK
#)
          @endnotes << wrap
        end
      end
      def json_head
        #metadata=SiSU_Metadata::Summary.new(@md).json.metadata
        #@per.head << metadata
      end
      def name_tags(dob)
        tags=''
        if defined? dob.tags \
        and dob.tags.length > 0 # insert tags "hypertargets"
          dob.tags.each do |t|
            tags=tags << %{<named id="#{t}" />}
          end
        end
        tags
      end
      def json_structure(dob,attrib=nil)
        if dob.is ==:para \
        || dob.is ==:heading
          if dob.is==:heading
            lv=dob.ln
            dob.ln + 2
          else lv=nil
          end
          extract_endnotes(dob)
          dob.obj=dob.obj.
            gsub(/#{Mx[:en_a_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_a_c]}/,'<en>\1</en>'). #footnote/endnote clean
            gsub(/#{Mx[:en_b_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_b_c]}/,'<en>\1</en>')
          util=SiSU_JSONutils::Clean.new(dob.obj)
          wrapped=util.line_json_clean
          @per.body << Ax[:tab]*1 + '{'
          if defined? dob.ocn and dob.ocn
            @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
          end
          if lv                                                                 # main text, contents, body KEEP
            @per.body <<
              Ax[:tab]*2 + %{"object": "<h#{lv}>} + wrapped + %{</h#{lv}>} +
              ((@endnotes.length > 0) ? '",' : '"')
          else
            @per.body <<
              Ax[:tab]*2 + '"object": "' + wrapped +
              ((@endnotes.length > 0) ? '",' : '"')
          end
          if @endnotes.length > 0                                               # main text, endnotes KEEP
            @per.body <<
              Ax[:tab]*2 + '"endnotes": "' +
              @endnotes.compact.join.strip + '"'
          end
          @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
          @endnotes=[]
        end
      end
      def block_structure(dob)
        dob=@trans.markup_block(dob)
        dob.obj=dob.obj.strip.
          gsub(/#{Mx[:en_a_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_a_c]}/,'<en>\1</en>'). #footnote/endnote clean
          gsub(/#{Mx[:en_b_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_b_c]}/,'<en>\1</en>') #footnote/endnote clean
        @per.body << Ax[:tab]*1 + '{'
        if defined? dob.ocn and dob.ocn
          @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
        end
        @per.body <<
          Ax[:tab]*2 + '"object": "' + dob.obj + '"'
          #((@endnotes.length > 0) ? '",' : '"')
        @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
      end
      def group_structure(dob)
        dob=@trans.markup_group(dob)
        dob.obj=dob.obj.strip.
          gsub(/#{Mx[:en_a_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_a_c]}/,'<en>\1</en>'). #footnote/endnote clean
          gsub(/#{Mx[:en_b_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_b_c]}/,'<en>\1</en>') #footnote/endnote clean
        @per.body << Ax[:tab]*1 + '{'
        if defined? dob.ocn and dob.ocn
          @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
        end
        @per.body <<
          Ax[:tab]*2 + '"object": "' + dob.obj + '"'
          #((@endnotes.length > 0) ? '",' : '"')
        @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
      end
      def poem_structure(dob)
        dob=@trans.markup_group(dob)
        dob.obj=dob.obj.strip
        @per.body << Ax[:tab]*1 + '{'
        if defined? dob.ocn and dob.ocn
          @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
        end
        @per.body <<
          Ax[:tab]*2 + '"object": "' + dob.obj + '"'
          #((@endnotes.length > 0) ? '",' : '"')
        @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
      end
      def code_structure(dob)
        dob=@trans.markup_group(dob)
        dob.obj=dob.obj.gsub(/\s\s/,'&#160;&#160;').strip
        @per.body << Ax[:tab]*1 + '{'
        if defined? dob.ocn and dob.ocn
          @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
        end
        @per.body <<
          Ax[:tab]*2 + '"object": "' + dob.obj + '"'
          #((@endnotes.length > 0) ? '",' : '"')
        @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
      end
      def table_structure(dob)
        table=SiSU_JSON_Shared::TableJSON.new(dob)
        @per.body << Ax[:tab]*1 + '{'
        if defined? dob.ocn and dob.ocn
          @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
        end
        @per.body <<
          Ax[:tab]*2 + '"object": "' + table.table.obj + '"'
        #((@endnotes.length > 0) ? '",' : '"')
        @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
      end
      def markup(data)
        @endnotes=[]
        @rcdc=false
        @level,@cont,@copen,@json_contents_close=[],[],[],[]
        json_head
        (0..7).each { |x| @cont[x]=@level[x]=false }
        (4..7).each { |x| @json_contents_close[x]='' }
        data.each_with_index do |dob,i|
          dob=@trans.char_enc.utf8(dob) if @sys.locale =~/utf-?8/i #% utf8
          dob=@trans.markup(dob)
          if @rcdc==false \
          and (dob.obj =~/~meta/ \
          and dob.obj =~/Document Information/)
            @rcdc=true
          end
          if dob.obj !~/(^#{Rx[:meta]}|#{Mx[:br_eof]}|#{Mx[:br_endnotes]})/
            if not @rcdc
              x=SiSU_JSON_Format::FormatTextObject.new(@md,dob)
              if dob.is==:heading
                json_structure(dob)
                dob.obj=case dob.ln
                when 0 then x.heading_body0
                when 1 then x.heading_body1
                when 2 then x.heading_body2
                when 3 then x.heading_body3
                when 4 then x.heading_body4
                when 5 then x.heading_body5
                when 6 then x.heading_body6
                when 7 then x.heading_body7
                end
              else
                if dob.is ==:verse
                  poem_structure(dob)
                elsif dob.is ==:group
                  group_structure(dob)
                elsif dob.is ==:block
                  block_structure(dob)
                elsif dob.is ==:code
                  code_structure(dob)
                elsif dob.is ==:table
                  table_structure(dob)
                elsif dob.is ==:para \
                and dob.indent.to_s =~/[1-9]/ \
                and dob.bullet_==true
                  json_structure(dob,"indent_bullet#{dob.indent}")
                elsif dob.is ==:para \
                and dob.indent.to_s =~/[1-9]/ \
                and dob.indent == dob.hang
                  json_structure(dob,"indent#{dob.indent}")
                elsif dob.is==:para \
                and dob.hang.to_s =~/[0-9]/ \
                and dob.indent != dob.hang
                  json_structure(dob,"hang#{dob.hang.to_s}_indent#{dob.indent.to_s}")
                else json_structure(dob)
                end
              end
            end
            dob.obj=dob.obj.gsub(/#{Mx[:pa_o]}:\S+#{Mx[:pa_c]}/,'') if dob.obj
          end
        end
        6.downto(4) do |x|
          y=x - 1; v=x - 3
          @per.body << "#{Ax[:tab]*5}</content>\n#{Ax[:tab]*y}</contents#{v}>" if @level[x]==true
        end
        3.downto(1) do |x|
          y=x - 1
          @per.body << "#{Ax[:tab]*y}</heading#{x}>" if @level[x]==true
        end
      end
      def pre
        @per.head,@per.body=[],[]
        @per.open = '{'
      end
      def post
        @per.close = '}'
      end
      def publish
        content=[]
        @per.body[-1] = @per.body[-1].gsub(/,$/, '') #= Ax[:tab]*1 + '}'
        content << @per.open << @per.head << @per.body << @per.metadata
        content << @per.tail << @per.close
        content=content.flatten.compact
        Output.new(content,@md).json
      end
    end
    class Output
      def initialize(data,md)
        @data,@md=data,md
        @file=SiSU_Env::FileOp.new(@md)
      end
      def json
        SiSU_Env::FileOp.new(@md).mkdir
        filename_json=@file.write_file.json
        @data.each do |str|
          str=str.gsub(/\A\s+\Z/m,'') #str.gsub(/^\s+$/,'')
          filename_json.puts str unless str.empty?
        end
        filename_json.close
      end
    end
  end
end
__END__

==> sisu_7.2.1.orig.tar.xz: sisu-7.2.1/lib/sisu/json.rb <==
#encoding: utf-8
#- Name: SiSU
#
#  - Description: documents, structuring, processing, publishing, search
#    json
#
#  - Author: Ralph Amissah
#    <ralph.amissah@gmail.com>
#
#  - Copyright: (C) 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006,
#    2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2019,
#    2020, 2021, Ralph Amissah,
#    All Rights Reserved.
#
#  - License: GPL 3 or later:
#
#    SiSU, a framework for document structuring, publishing and search
#
#    Copyright (C) Ralph Amissah
#
#    This program is free software: you can redistribute it and/or modify it
#    under the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    This program is distributed in the hope that it will be useful, but WITHOUT
#    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
#    more details.
#
#    You should have received a copy of the GNU General Public License along with
#    this program. If not, see <http://www.gnu.org/licenses/>.
#
#    If you have Internet connection, the latest version of the GPL should be
#    available at these locations:
#    <http://www.fsf.org/licensing/licenses/gpl.html>
#    <http://www.gnu.org/licenses/gpl.html>
#
#    <http://www.sisudoc.org/sisu/en/manifest/gpl.fsf.html>
#
#  - SiSU uses:
#    - Standard SiSU markup syntax,
#    - Standard SiSU meta-markup syntax, and the
#    - Standard SiSU object citation numbering and system
#
#  - Homepages:
#    <http://www.sisudoc.org>
#
#  - Git
#    <https://git.sisudoc.org/projects/>
#    <https://git.sisudoc.org/projects/?p=software/sisu.git;a=summary>
#    <https://git.sisudoc.org/projects/?p=markup/sisu-markup-samples.git;a=summary>
module SiSU_JSON
  require_relative 'se_hub_particulars'                 # se_hub_particulars.rb
    include SiSU_Particulars
  require_relative 'se'                                 # se.rb
    include SiSU_Env
  require_relative 'json_shared'                        # json_shared.rb
    include SiSU_JSON_Munge
  require_relative 'json_format'                        # json_format.rb
    include SiSU_JSON_Format
  require_relative 'json_persist'                       # json_persist.rb
  require_relative 'shared_metadata'                    # shared_metadata.rb
  @@alt_id_count=0
  @@tablefoot=''
  class Source
    def initialize(opt)
      @opt=opt
      @particulars=SiSU_Particulars::CombinedSingleton.instance.get_all(opt)
    end
    def read
      begin
        @env,@md,@ao_array=@particulars.env,@particulars.md,@particulars.ao_array
        unless @opt.act[:quiet][:set]==:on
          tool=if (@opt.act[:verbose_plus][:set]==:on \
          || @opt.act[:maintenance][:set]==:on)
            @env.program.web_browser +
            ' file://' +
            @md.file.output_path.json.dir + '/' +
            @md.file.base_filename.json
          elsif @opt.act[:verbose][:set]==:on
            @env.program.web_browser +
            ' file://' +
            @md.file.output_path.json.dir + '/' +
            @md.file.base_filename.json
          else "[#{@opt.f_pth[:lng_is]}] #{@opt.fno}"
          end
          (@opt.act[:verbose][:set]==:on \
          || @opt.act[:verbose_plus][:set]==:on \
          || @opt.act[:maintenance][:set]==:on) \
          ? SiSU_Screen::Ansi.new(
              @opt.act[:color_state][:set],
              'JSON',
              tool
            ).green_hi_blue
          : SiSU_Screen::Ansi.new(
              @opt.act[:color_state][:set],
              'JSON',
              tool
            ).green_title_hi
          if (@opt.act[:verbose_plus][:set]==:on \
          || @opt.act[:maintenance][:set]==:on)
            SiSU_Screen::Ansi.new(
              @opt.act[:color_state][:set],
              @opt.fns,
                '/' + @md.file.output_path.json.dir +
                '/' + @md.file.base_filename.json
            ).flow
          end
        end
        SiSU_JSON::Source::Songsheet.new(@particulars).song
      rescue
        SiSU_Errors::Rescued.new($!,$@,@opt.selections.str,@opt.fns).location do
          __LINE__.to_s + ':' + __FILE__
        end
      ensure
        SiSU_Env::CreateSite.new(@opt).cp_css
        Dir.chdir(@opt.f_pth[:pth])
      end
    end
    private
    class Songsheet
      def initialize(particulars)
        @env,@md,@ao_array,@particulars=
          particulars.env,particulars.md,particulars.ao_array,particulars
        @file=SiSU_Env::FileOp.new(@md)
      end
      def song
        begin
          SiSU_JSON::Source::Scroll.new(@particulars).songsheet
        rescue
          SiSU_Errors::Rescued.new($!,$@,@md.opt.selections.str,@md.fns).location do
            __LINE__.to_s + ':' + __FILE__
          end
        ensure
        end
      end
    end
    class Scroll
      require_relative 'json_shared'                   # json_shared.rb #check already called
      require_relative 'txt_shared'                     # txt_shared.rb
        include SiSU_TextUtils
      require_relative 'css'                            # css.rb
      def initialize(particulars)
        @env,@md,@ao_array=particulars.env,particulars.md,particulars.ao_array
        @tab="\t"
        @trans=SiSU_JSON_Munge::Trans.new(@md)
        @sys=SiSU_Env::SystemCall.new
        @per=SiSU_JSON_Persist::Persist.new
      end
      def songsheet
        begin
          pre
          @data=markup(@ao_array)
          post
          publish
        ensure
          SiSU_JSON_Persist::Persist.new.persist_init
        end
      end
    protected
      def embedded_endnotes(dob='')
        dob.obj=dob.obj.gsub(/#{Mx[:en_a_o]}(\d+)\s+(.+?)#{Mx[:en_a_c]}/,
            '<endnote><number>\1</number><note>\2</note></endnote> ').
          gsub(/#{Mx[:en_b_o]}([*+]\d+)\s+(.+?)#{Mx[:en_b_c]}/,
            '<endnote><symbol>\1</symbol><note>\2</note></endnote> ').
          gsub(/#{Mx[:en_a_o]}([*+]+)\s+(.+?)#{Mx[:en_a_c]}/,
            '<endnote><symbol>\1</symbol><note>\2</note></endnote> ')
      end
      def extract_endnotes(dob='')
        notes=dob.obj.scan(/(?:#{Mx[:en_a_o]}|#{Mx[:en_b_o]})([\d*+]+\s+.+?)(?:#{Mx[:en_a_c]}|#{Mx[:en_b_c]})/)
        notes.flatten.each do |e|
          s=e.to_s
          util=SiSU_JSONutils::Clean.new(s)
          wrap=util.line_json_clean
          wrap=wrap.gsub(/^(\d+)\s+(.+?)\s*\Z/m, <<-WOK
\\n[\\1.] \\2
              WOK
            ).
            gsub(/^([*+]\d+)\s+(.+?)\s*\Z/m, <<-WOK
\\n[\\1.] \\2
              WOK
            ).
            gsub(/^([*+]+)\s+(.+?)\s*\Z/m, <<-WOK
\\n[\\1.] \\2
              WOK
            ).strip
#KEEP alternative presentation of endnotes
#        wrap=wrap.gsub(/^(\d+)\s+(.+?)\s*\Z/m, <<WOK
##{Ax[:tab]*1}<p class="endnote" notenumber="\\1">
##{Ax[:tab]*2}\\1. \\2
##{Ax[:tab]*1}</p>
#WOK
#)
          @endnotes << wrap
        end
      end
      def json_head
        #metadata=SiSU_Metadata::Summary.new(@md).json.metadata
        #@per.head << metadata
      end
      def name_tags(dob)
        tags=''
        if defined? dob.tags \
        and dob.tags.length > 0 # insert tags "hypertargets"
          dob.tags.each do |t|
            tags=tags << %{<named id="#{t}" />}
          end
        end
        tags
      end
      def json_structure(dob,attrib=nil)
        if dob.is ==:para \
        || dob.is ==:heading
          if dob.is==:heading
            lv=dob.ln
            dob.ln + 2
          else lv=nil
          end
          extract_endnotes(dob)
          dob.obj=dob.obj.
            gsub(/#{Mx[:en_a_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_a_c]}/,'<en>\1</en>'). #footnote/endnote clean
            gsub(/#{Mx[:en_b_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_b_c]}/,'<en>\1</en>')
          util=SiSU_JSONutils::Clean.new(dob.obj)
          wrapped=util.line_json_clean
          @per.body << Ax[:tab]*1 + '{'
          if defined? dob.ocn and dob.ocn
            @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
          end
          if lv                                                                 # main text, contents, body KEEP
            @per.body <<
              Ax[:tab]*2 + %{"object": "<h#{lv}>} + wrapped + %{</h#{lv}>} +
              ((@endnotes.length > 0) ? '",' : '"')
          else
            @per.body <<
              Ax[:tab]*2 + '"object": "' + wrapped +
              ((@endnotes.length > 0) ? '",' : '"')
          end
          if @endnotes.length > 0                                               # main text, endnotes KEEP
            @per.body <<
              Ax[:tab]*2 + '"endnotes": "' +
              @endnotes.compact.join.strip + '"'
          end
          @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
          @endnotes=[]
        end
      end
      def block_structure(dob)
        dob=@trans.markup_block(dob)
        dob.obj=dob.obj.strip.
          gsub(/#{Mx[:en_a_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_a_c]}/,'<en>\1</en>'). #footnote/endnote clean
          gsub(/#{Mx[:en_b_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_b_c]}/,'<en>\1</en>') #footnote/endnote clean
        @per.body << Ax[:tab]*1 + '{'
        if defined? dob.ocn and dob.ocn
          @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
        end
        @per.body <<
          Ax[:tab]*2 + '"object": "' + dob.obj + '"'
          #((@endnotes.length > 0) ? '",' : '"')
        @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
      end
      def group_structure(dob)
        dob=@trans.markup_group(dob)
        dob.obj=dob.obj.strip.
          gsub(/#{Mx[:en_a_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_a_c]}/,'<en>\1</en>'). #footnote/endnote clean
          gsub(/#{Mx[:en_b_o]}([\d*+]+)\s+(?:.+?)#{Mx[:en_b_c]}/,'<en>\1</en>') #footnote/endnote clean
        @per.body << Ax[:tab]*1 + '{'
        if defined? dob.ocn and dob.ocn
          @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
        end
        @per.body <<
          Ax[:tab]*2 + '"object": "' + dob.obj + '"'
          #((@endnotes.length > 0) ? '",' : '"')
        @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
      end
      def poem_structure(dob)
        dob=@trans.markup_group(dob)
        dob.obj=dob.obj.strip
        @per.body << Ax[:tab]*1 + '{'
        if defined? dob.ocn and dob.ocn
          @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
        end
        @per.body <<
          Ax[:tab]*2 + '"object": "' + dob.obj + '"'
          #((@endnotes.length > 0) ? '",' : '"')
        @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
      end
      def code_structure(dob)
        dob=@trans.markup_group(dob)
        dob.obj=dob.obj.gsub(/\s\s/,'&#160;&#160;').strip
        @per.body << Ax[:tab]*1 + '{'
        if defined? dob.ocn and dob.ocn
          @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
        end
        @per.body <<
          Ax[:tab]*2 + '"object": "' + dob.obj + '"'
          #((@endnotes.length > 0) ? '",' : '"')
        @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
      end
      def table_structure(dob)
        table=SiSU_JSON_Shared::TableJSON.new(dob)
        @per.body << Ax[:tab]*1 + '{'
        if defined? dob.ocn and dob.ocn
          @per.body << Ax[:tab]*2 + '"ocn": ' + dob.ocn.to_s + '",'
        end
        @per.body <<
          Ax[:tab]*2 + '"object": "' + table.table.obj + '"'
        #((@endnotes.length > 0) ? '",' : '"')
        @per.body << Ax[:tab]*1 + '},' # unless is last object then '}'
      end
      def markup(data)
        @endnotes=[]
        @rcdc=false
        @level,@cont,@copen,@json_contents_close=[],[],[],[]
        json_head
        (0..7).each { |x| @cont[x]=@level[x]=false }
        (4..7).each { |x| @json_contents_close[x]='' }
        data.each_with_index do |dob,i|
          dob=@trans.char_enc.utf8(dob) if @sys.locale =~/utf-?8/i #% utf8
          dob=@trans.markup(dob)
          if @rcdc==false \
          and (dob.obj =~/~meta/ \
          and dob.obj =~/Document Information/)
            @rcdc=true
          end
          if dob.obj !~/(^#{Rx[:meta]}|#{Mx[:br_eof]}|#{Mx[:br_endnotes]})/
            if not @rcdc
              x=SiSU_JSON_Format::FormatTextObject.new(@md,dob)
              if dob.is==:heading
                json_structure(dob)
                dob.obj=case dob.ln
                when 0 then x.heading_body0
                when 1 then x.heading_body1
                when 2 then x.heading_body2
                when 3 then x.heading_body3
                when 4 then x.heading_body4
                when 5 then x.heading_body5
                when 6 then x.heading_body6
                when 7 then x.heading_body7
                end
              else
                if dob.is ==:verse
                  poem_structure(dob)
                elsif dob.is ==:group
                  group_structure(dob)
                elsif dob.is ==:block
                  block_structure(dob)
                elsif dob.is ==:code
                  code_structure(dob)
                elsif dob.is ==:table
                  table_structure(dob)
                elsif dob.is ==:para \
                and dob.indent.to_s =~/[1-9]/ \
                and dob.bullet_==true
                  json_structure(dob,"indent_bullet#{dob.indent}")
                elsif dob.is ==:para \
                and dob.indent.to_s =~/[1-9]/ \
                and dob.indent == dob.hang
                  json_structure(dob,"indent#{dob.indent}")
                elsif dob.is==:para \
                and dob.hang.to_s =~/[0-9]/ \
                and dob.indent != dob.hang
                  json_structure(dob,"hang#{dob.hang.to_s}_indent#{dob.indent.to_s}")
                else json_structure(dob)
                end
              end
            end
            dob.obj=dob.obj.gsub(/#{Mx[:pa_o]}:\S+#{Mx[:pa_c]}/,'') if dob.obj
          end
        end
        6.downto(4) do |x|
          y=x - 1; v=x - 3
          @per.body << "#{Ax[:tab]*5}</content>\n#{Ax[:tab]*y}</contents#{v}>" if @level[x]==true
        end
        3.downto(1) do |x|
          y=x - 1
          @per.body << "#{Ax[:tab]*y}</heading#{x}>" if @level[x]==true
        end
      end
      def pre
        @per.head,@per.body=[],[]
        @per.open = '{'
      end
      def post
        @per.close = '}'
      end
      def publish
        content=[]
        @per.body[-1] = @per.body[-1].gsub(/,$/, '') #= Ax[:tab]*1 + '}'
        content << @per.open << @per.head << @per.body << @per.metadata
        content << @per.tail << @per.close
        content=content.flatten.compact
        Output.new(content,@md).json
      end
    end
    class Output
      def initialize(data,md)
        @data,@md=data,md
        @file=SiSU_Env::FileOp.new(@md)
      end
      def json
        SiSU_Env::FileOp.new(@md).mkdir
        filename_json=@file.write_file.json
        @data.each do |str|
          str=str.gsub(/\A\s+\Z/m,'') #str.gsub(/^\s+$/,'')
          filename_json.puts str unless str.empty?
        end
        filename_json.close
      end
    end
  end
end
__END__
